#include "pdfpageitem.h"
#include "verticalview.h"
#include "sgtileitem.h"
#include "pdfrendertask.h"
#include "pdfdocument.h"
#include "pdfzoom.h"
#include "twips.h"
#include "pagedecoration.h"
#include "pageoverlay.h"

#include "config.h"

#include <QSharedPointer>
#include <QtCore/qmath.h>

#include <QQmlComponent>
#include <QQmlEngine>

PdfPageItem::PdfPageItem(VerticalView* view, QSharedPointer<PdfDocument> d)
    : QQuickItem(view)
    , m_view(view)
    , m_document(d)
{
    connect(m_view, &VerticalView::flickableResized, this, &PdfPageItem::updateVisibleRect);
    connect(m_view, &VerticalView::flickableScrolled, this, &PdfPageItem::updateVisibleRect);
    connect(m_view->zoomSettings(), &PdfZoom::zoomFactorChanged,
            this, &PdfPageItem::updateVisibleRect);

    connect(m_view, &VerticalView::cacheBufferChanged, this, &PdfPageItem::updateVisibleRect);

    connect(RenderEngine::instance(), &RenderEngine::taskRenderFinished,
            this, &PdfPageItem::updateTileData);

    addPageDecoration();
    addPageOverlay();
}

PdfPageItem::~PdfPageItem()
{    
    disconnect(RenderEngine::instance(), &RenderEngine::taskRenderFinished,
            this, &PdfPageItem::updateTileData);

    // Remove all tasks from rendering queue.
    for (auto i = m_tiles.begin(); i != m_tiles.end(); ++i)
        RenderEngine::instance()->dequeueTask(i.value()->id());

    qDeleteAll(m_tiles);

    delete m_decoration;
    delete m_overlay;
}

void PdfPageItem::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
{
    QQuickItem::geometryChanged(newGeometry, oldGeometry);
    addPageDecoration();
    addPageOverlay();
    updateVisibleRect();
}

void PdfPageItem::updateVisibleRect()
{
    const QRect &pageRect = QRect(x(), y(), width(), height());
    bool isInBufferArea = pageRect.intersects(m_view->bufferArea());

    if (!m_tiles.isEmpty()) {
        auto i = m_tiles.begin();
        while (i != m_tiles.end()) {
            SGTileItem* sgtile = i.value();
            const QRect &tileAreaInView = this->mapRectToItem(m_view, sgtile->area()).toRect();

            // Ok - we still need this item.
            if (m_view->bufferArea().intersects(tileAreaInView)) {
                i++;
                continue;
            }

            RenderEngine::instance()->dequeueTask(sgtile->id());
            sgtile->deleteLater();
            i = m_tiles.erase(i);
        }
    }

    if (isInBufferArea) {
        // Number of tiles per row
        int tilesPerWidth           = qCeil(this->width()  / TILE_SIZE);
        int tilesPerHeight          = qCeil(this->height() / TILE_SIZE);

        // Get indexes for visible tiles
        int visiblesFromWidth       = 0;
        int visiblesFromHeight      = 0;
        int visiblesToWidth         = qCeil(qreal(this->width()) / TILE_SIZE);
        int visiblesToHeight        = qCeil(qreal(this->height()) / TILE_SIZE);

        this->generateTiles(visiblesFromWidth, visiblesFromHeight, visiblesToWidth, visiblesToHeight, tilesPerWidth, tilesPerHeight);
    }
}

void PdfPageItem::generateTiles(int x1, int y1, int x2, int y2, int tilesPerWidth, int tilesPerHeight)
{
    for (int x = x1; x < x2; x++) {
        for (int y = y1; y < y2; y++) {
            bool lastRow    = (y == (tilesPerHeight - 1));
            bool lastColumn = (x == (tilesPerWidth  - 1));

            int left   = TILE_SIZE * x;
            int top    = TILE_SIZE * y;
            int width  = lastColumn ? this->width() - left : TILE_SIZE;
            int height = lastRow    ? this->height() - top : TILE_SIZE;

            QRect tileRect(left, top, width, height);
            int index = x + tilesPerWidth * y;

            createTile(index, tileRect);
        }
    }
}

void PdfPageItem::createTile(int index, const QRect &rect)
{
    if (!m_tiles.contains(index)) {
        auto tile = new SGTileItem(rect, m_view->zoomSettings()->zoomFactor(), RenderEngine::getNextId(), this);
        m_tiles.insert(index, tile);

        RenderEngine::instance()->enqueueTask(createTask(rect, tile->id()));
    }
}

void PdfPageItem::clearView()
{
    for (auto i = m_tiles.begin(); i != m_tiles.end(); ++i)
        RenderEngine::instance()->dequeueTask(i.value()->id());

    auto i = m_tiles.begin();
    while (i != m_tiles.end()) {
        SGTileItem* sgtile = i.value();
        sgtile->deleteLater();
        i = m_tiles.erase(i);
    }
}

void PdfPageItem::addPageDecoration()
{
    m_decoration = new PageDecoration(this);

    m_decoration->setZ(-1000);  // stay below everything
    m_decoration->setWidth(this->width());
    m_decoration->setHeight(this->height());
}

void PdfPageItem::addPageOverlay()
{
    m_overlay = new PageOverlay(m_view, m_index, this);

    m_overlay->setZ(1000);  // stay above everything
    m_overlay->setWidth(this->width());
    m_overlay->setHeight(this->height());
}

PdfRenderTask* PdfPageItem::createTask(const QRect &rect, int id) const
{
    auto task = new PdfRenderTask();
    task->setId(id);
    task->setDocument(m_document);
    task->setPage(m_index);
    task->setArea(rect);
    task->setZoom(m_view->zoomSettings()->zoomFactor());
    return task;
}

void PdfPageItem::updateTileData(AbstractRenderTask *task, QImage img)
{
    int id = task->id();
    for (auto i = m_tiles.begin(); i != m_tiles.end(); ++i) {
        SGTileItem* sgtile = i.value();
        if (sgtile->id() == id) {
            sgtile->setData(img);
            break;
        }
    }
}
