#include "verticalview.h"
#include "pdfdocument.h"
#include "pdfpageitem.h"
#include "pdfzoom.h"
#include "twips.h"

#include <QImage>
#include <QtCore/qmath.h>

// FIXME: Sometimes layout is not created when a document is opened
// TODO: Check if everything works as expected when zooming

#define UPDATE_TIMER_MSECS 20

VerticalView::VerticalView(QQuickItem *parent)
    : QQuickItem(parent)
    , m_parentFlickable(nullptr)
    , m_document(nullptr)
    , m_zoomSettings(new PdfZoom(this))
    , m_cacheBuffer(200)
    , m_spacing(12)
    , m_currentPageIndex(-1)
    , m_visibleArea(0, 0, 0, 0)
    , m_bufferArea(0, 0, 0, 0)
    , m_renderHints(0)
    , m_showLinkHighlight(true)
    , m_linkHighlightColor(Qt::red)
    , m_error(PopplerError::NoError)
    , m_hasZoomChanged(false)
    , m_hasFlickableBeenScrolled(false)
    , m_hasFlickableBeenResized(false)
    , m_hasRenderHintsChanged(false)
{
    // Connect signals
    connect(this, &VerticalView::documentChanged, this, &VerticalView::updateLayout);
    connect(this, &VerticalView::parentFlickableChanged, this, &VerticalView::updateLayout);
    connect(this, &VerticalView::cacheBufferChanged, this, &VerticalView::updateLayout);
    connect(this, &VerticalView::spacingChanged, this, &VerticalView::updateLayout);
    connect(this, &VerticalView::renderHintsChanged, this, &VerticalView::onRenderHintsChanged);

    connect(&m_updateTimer, &QTimer::timeout, this, &VerticalView::updateLayout);

    connect(m_zoomSettings, &PdfZoom::zoomFactorChanged, [&]() {
        m_hasZoomChanged = true;
        updateLayout();
    });
}

VerticalView::~VerticalView()
{
    qDeleteAll(m_pages);
}

// Returns the parent QML Flickable
QQuickItem* VerticalView::parentFlickable() const
{
    return m_parentFlickable;
}

// Set the parent QML Flickable
void VerticalView::setParentFlickable(QQuickItem *flickable)
{
    if (m_parentFlickable == flickable)
        return;

    if (m_parentFlickable)
        m_parentFlickable->disconnect(this);

    m_parentFlickable = flickable;

    connect(m_parentFlickable, &QQuickItem::widthChanged, this, &VerticalView::onFlickableSizeChanged);
    connect(m_parentFlickable, &QQuickItem::heightChanged, this, &VerticalView::onFlickableSizeChanged);

    connect(m_parentFlickable, SIGNAL(contentXChanged()), this, SLOT(onFlickableScrolled()));
    connect(m_parentFlickable, SIGNAL(contentYChanged()), this, SLOT(onFlickableScrolled()));

    Q_EMIT parentFlickableChanged();
}

PdfDocument* VerticalView::document() const
{
    return m_document.data();
}

PdfZoom *VerticalView::zoomSettings() const
{
    return m_zoomSettings;
}
void VerticalView::setCurrentPageIndex(int currentPageIndex)
{
    if (m_currentPageIndex == currentPageIndex)
        return;

    m_currentPageIndex = currentPageIndex;
    Q_EMIT currentPageIndexChanged();
}

int VerticalView::cacheBuffer() const
{
    return m_cacheBuffer;
}

void VerticalView::setCacheBuffer(int cacheBuffer)
{
    if (m_cacheBuffer == cacheBuffer)
        return;

    m_cacheBuffer = cacheBuffer;
    Q_EMIT cacheBufferChanged();
}

int VerticalView::spacing() const
{
    return m_spacing;
}

void VerticalView::setSpacing(int spacing)
{
    if (m_spacing == spacing)
        return;

    m_spacing = spacing;
    Q_EMIT spacingChanged();
}

int VerticalView::currentPageIndex() const
{
    return m_currentPageIndex;
}

int VerticalView::pagesCount() const
{
    if (!m_document || m_document.data()->isLocked())
        return 0;

    return m_document.data()->pageCount();
}

PopplerError::Error VerticalView::error() const
{
    return m_error;
}

QRect VerticalView::visibleArea() const
{
    return m_visibleArea;
}

QRect VerticalView::bufferArea() const
{
    return m_bufferArea;
}

PdfDocument::RenderHints VerticalView::renderHints() const
{
    return m_renderHints;
}

void VerticalView::setRenderHints(const PdfDocument::RenderHints hints)
{
    if (m_renderHints == hints)
        return;

    m_renderHints = hints;
    Q_EMIT renderHintsChanged();
}

bool VerticalView::showLinkHighlight() const
{
    return m_showLinkHighlight;
}

void VerticalView::setShowLinkHighlight(bool show)
{
    if (m_showLinkHighlight == show)
        return;

    m_showLinkHighlight = show;
    Q_EMIT showLinkHighlightChanged();
}

QColor VerticalView::linkHighlightColor() const
{
    return m_linkHighlightColor;
}

void VerticalView::setLinkHighlightColor(const QColor &color)
{
    if (m_linkHighlightColor == color)
        return;

    m_linkHighlightColor = color;
    Q_EMIT linkHighlightColorChanged();
}

void VerticalView::setError(const PopplerError::Error &error)
{
    if (m_error == error)
        return;

    m_error = error;
    Q_EMIT errorChanged();
}

void VerticalView::initializeDocument(const QString &path)
{
    if (m_document)
        m_document->disconnect(this);

    setError(PopplerError::NoError);

    m_document = QSharedPointer<PdfDocument>(new PdfDocument());
    m_document->setPath(path);

    /* A lot of things happens when we set the path property in
     * m_document. Need to check if an error has been emitted. */
    setError(m_document->error());

    if (m_error != PopplerError::NoError && m_error != PopplerError::DocumentLocked) {
        m_document.clear();

        // Stop doing anything below.
        return;
    }

    m_document.data()->setRenderHints(m_renderHints);

    Q_EMIT documentChanged();

    // Init zoom settings
    m_zoomSettings->init();
}

bool VerticalView::adjustZoomToWidth()
{
    if (!m_zoomSettings)
        return false;

    bool result;

    result = m_zoomSettings->adjustZoomToWidth();
    m_hasZoomChanged = result;

    return result;
}

bool VerticalView::adjustZoomToPage()
{
    if (!m_zoomSettings)
        return false;

    bool result;

    result = m_zoomSettings->adjustZoomToPage();
    m_hasZoomChanged = result;

    return result;
}

bool VerticalView::adjustAutomaticZoom()
{
    if (!m_zoomSettings)
        return false;

    bool result;

    result = m_zoomSettings->adjustAutomaticZoom();
    m_hasZoomChanged = result;

    return result;
}

void VerticalView::positionAtBeginning()
{
    m_parentFlickable->setProperty("contentY", 0);
}

void VerticalView::positionAtIndex(int index, qreal top, qreal left)
{
    PdfPageItem* page = m_pages.value(index);

    m_parentFlickable->setProperty("contentY", page->x() + left);
    m_parentFlickable->setProperty("contentY", page->y() + top);
}

void VerticalView::positionAtEnd()
{
    const int posY = this->height() - m_parentFlickable->property("height").toInt();

    m_parentFlickable->setProperty("contentY", posY);
}

QVariantMap VerticalView::linkAtPosition(int x, int y)
{
    QVariantMap result;
    int pageIndex = -1;
    PdfPageItem* page;

    for (int i=0; i<m_pages.count(); ++i) {
        page = m_pages.value(i);

        if (QRect(page->x(), page->y(), page->width(), page->height()).contains(x, y)) {
            pageIndex = i;
            break;
        }
    }

    if (pageIndex == -1) {
        return QVariantMap();
    }

    QList<Poppler::Link *> links = m_document->pageLinks(pageIndex);
    Q_FOREACH (Poppler::Link *link, links) {

        QRectF linkRect(link->linkArea().x() * page->width(),
                        link->linkArea().y() * page->height(),
                        link->linkArea().width() * page->width(),
                        link->linkArea().height() * page->height());

        if (linkRect.contains(x - page->x(), y - page->y())) {
            if (link->linkType() == Poppler::Link::LinkType::Goto) {
                auto l = static_cast<Poppler::LinkGoto*>(link);

                if (!l->isExternal()) {
                    Poppler::LinkDestination dest = l->destination();

                    result["pageIndex"] = dest.pageNumber() - 1;
                    result["top"] = dest.top();
                    result["left"] = dest.left();

                    break;
                }
            } else if (link->linkType() == Poppler::Link::LinkType::Browse) {
                auto l = static_cast<Poppler::LinkBrowse*>(link);

                result["url"] = l->url();

                break;
            }
        }
    }

    return result;
}

bool VerticalView::unlock(const QString &ownerPassword, const QString &userPassword)
{
    bool result = false;

    if (m_document) {
        result = m_document.data()->unlock(ownerPassword, userPassword);

        if (result) {
            Q_EMIT documentChanged(); // which trigger a layout update
        }
    }

    return result;
}

void VerticalView::updateLayout()
{
    if (!m_parentFlickable || !m_document || m_document.data()->isLocked())
        return;

    if (m_zoomSettings->zoomMode() == PdfZoom::FitWidth) {
        if (m_zoomSettings->adjustZoomToWidth(false))
            m_hasZoomChanged = true;
    }

    else if (m_zoomSettings->zoomMode() == PdfZoom::FitPage) {
        if (m_zoomSettings->adjustZoomToPage(false))
            m_hasZoomChanged = true;
    }

    else if (m_zoomSettings->zoomMode() == PdfZoom::Automatic) {
        if (m_zoomSettings->adjustAutomaticZoom(false))
            m_hasZoomChanged = true;
    }

    if (!m_pages.isEmpty() && (m_hasZoomChanged || m_hasRenderHintsChanged)) {
        clearView();
        m_hasZoomChanged = false;
        m_hasRenderHintsChanged = false;
    }

    // Just for convenience.
    QRect documentRect(this->boundingRect().toRect());

    // Update visible area
    QRect visibleRect(m_parentFlickable->property("contentX").toInt(),
                      m_parentFlickable->property("contentY").toInt(),
                      m_parentFlickable->width(),
                      m_parentFlickable->height());

    m_visibleArea = visibleRect.intersected(documentRect);

    // Update buffer area
    QRect bufferRect(m_visibleArea.left()   -  m_cacheBuffer,
                     m_visibleArea.top()    -  m_cacheBuffer,
                     m_visibleArea.width()  + (m_cacheBuffer * 2),
                     m_visibleArea.height() + (m_cacheBuffer * 2));

    m_bufferArea = bufferRect.intersected(documentRect);

    PdfDocument* doc = m_document.data();

    const int totalSpacingPxs = m_spacing * (m_document.data()->pageCount() + 1);

    const qreal &zoomFactor = m_zoomSettings->zoomFactor();

    setWidth(Twips::convertPointsToPixels(doc->documentSize().width(), zoomFactor));
    setHeight(Twips::convertPointsToPixels(doc->documentSize().height(), zoomFactor) + totalSpacingPxs);

    for (int j=0; j < doc->pageCount(); ++j) {
        createPage(j);
    }

    if (m_hasFlickableBeenResized) {
        Q_EMIT flickableResized();
        m_hasFlickableBeenResized = false;
    }

    if (m_hasFlickableBeenScrolled) {
        Q_EMIT flickableScrolled();
        m_hasFlickableBeenScrolled = false;
    }
}

void VerticalView::createPage(int index)
{
    PdfPageItem* page;

    if (!m_pages.contains(index)) {
        page = new PdfPageItem(this, m_document);
        page->setIndex(index);

        // Set size
        const QSize &pageSize = m_document.data()->pageSize(index);
        const qreal &zoomFactor = m_zoomSettings->zoomFactor();

        page->setWidth(Twips::convertPointsToPixels(pageSize.width(), zoomFactor));
        page->setHeight(Twips::convertPointsToPixels(pageSize.height(), zoomFactor));

        m_pages.insert(index, page);
    } else {
        page = m_pages.value(index);
    }

    // Set X-pos - align to horizontal center
    page->setX((this->width() - page->width()) * 0.5);

    // Set Y-pos
    if (index == 0) {
        page->setY(m_spacing);
    } else {
        PdfPageItem* prevPage = m_pages.value(index - 1);
        page->setY(prevPage->y() + prevPage->height() + m_spacing);
    }
}

void VerticalView::clearView()
{
    auto i = m_pages.begin();
    while (i != m_pages.end()) {
        PdfPageItem* page = i.value();
        page->deleteLater();
        i = m_pages.erase(i);
    }
}

void VerticalView::onFlickableScrolled()
{
    if (m_updateTimer.isActive())
        return;

    m_hasFlickableBeenScrolled = true;

    // Update current page index
    int newCurrentPageIndex = -1;
    int flickCenterY = m_parentFlickable->property("contentY").toInt() + m_parentFlickable->height() * 0.5;

    // First attempt: check at the exact center
    Q_FOREACH(PdfPageItem* page, m_pages) {
        QRect pageRect(page->x(), page->y(), page->width(), page->height());
        QPoint flickableCenter(this->width() * 0.5, flickCenterY);

        if (pageRect.contains(flickableCenter)) {
            newCurrentPageIndex = page->index();
            break;
        }
    }

    // currentPageIndex not found
    // Second attempt: sum the spacing to flickCenterY, so we're sure we'll find a page this time.
    if (newCurrentPageIndex == -1) {
        flickCenterY += m_spacing;

        Q_FOREACH(PdfPageItem* page, m_pages) {
            QRect pageRect(page->x(), page->y(), page->width(), page->height());
            QPoint flickableCenter(this->width() * 0.5, flickCenterY);

            if (pageRect.contains(flickableCenter)) {
                newCurrentPageIndex = page->index();
                break;
            }
        }
    }

    // Set the value: -1 = current page not found
    setCurrentPageIndex(newCurrentPageIndex);

    // updateLayout() is triggered by the timer
    m_updateTimer.setSingleShot(true);
    m_updateTimer.start(UPDATE_TIMER_MSECS);
}

void VerticalView::onFlickableSizeChanged()
{
    m_hasFlickableBeenResized = true;

    updateLayout();
}

void VerticalView::onRenderHintsChanged()
{
    if (!m_document || m_document.data()->isLocked())
        return;

    m_hasRenderHintsChanged = true;

    m_document.data()->setRenderHints(m_renderHints);

    updateLayout();
}
